#include <stdio.h>
#include <stdlib.h>
#include "lrz.h"

// Fine-tuner. This might be able to be set to 16 without issue.
#define MAX_BITS 15

// Node type constants
#define NODE_UNKNOWN   0
#define NODE_BRANCHING 1
#define NODE_VALUE     2

// Decoder state constants
#define STATE_RAW  0
#define STATE_COPY 1

// Linked list data structure for binary tree nodes
typedef struct LRZ_NODE LRZ_NODE;
struct  LRZ_NODE {
    int type, value, done;
    LRZ_NODE *parent, *child[2];
};

// Atomic data processing unit for file chunks
typedef struct {

    // Input/output properties
    unsigned char *src, *dst;
    int srclen, dstlen, srcoff, dstoff;

    // Binary trees
    LRZ_NODE *rawtree, *lentree, *disttree;

    // Bit processing properties
    int bitsleft, reg, subbits, state;
} LRZ_CHUNK;


// Maximum bit value given MAX_BITS
#define CLAMP ((1 << MAX_BITS) - 1)



////////////////////////////////////////////////////////////////////////////////
//                             Decoder Functions                              //
////////////////////////////////////////////////////////////////////////////////

// This macro is used in lrzdFreeTree() to handle deletion of individual nodes
#define DeleteNode() {                                                  \
    if (node == node->parent->child[0]) node->parent->child[0] = NULL;  \
    if (node == node->parent->child[1]) node->parent->child[1] = NULL;  \
    temp = node->parent; free(node); node = temp; continue;             \
}

// Deletes memory allocated for a binary tree object
static void lrzdFreeTree(LRZ_NODE *root) {
    LRZ_NODE *temp, *node = root;

    // Exit now if the tree wasn't initialized
    if (root == NULL) return;

    // Traverse the tree until all child nodes are deleted
    while (1) {

        // Terminal nodes are deleted outright
        if (node->type != NODE_BRANCHING) DeleteNode();

        // Delete a branching node if both its children are gone
        if (node->child[0] == NULL && node->child[1] == NULL) {
            if (node == root) break; // Root has no parent, so exit here
            DeleteNode();
        }

        // Select the next non-deleted child node
        if (node->child[0] != NULL)
            node = node->child[0];
        else node = node->child[1];
    }

    // Delete root and return
    free(root);
    return;
}

// Deletes memory allocated for a chunk object
static void lrzdFreeChunk(LRZ_CHUNK *chunk) {
    lrzdFreeTree(chunk->rawtree);
    lrzdFreeTree(chunk->lentree);
    lrzdFreeTree(chunk->disttree);
    return;
}

// Decodes a binary tree from a chunk header
static int lrzdGetTree(LRZ_CHUNK *chunk, LRZ_NODE **out) {
    int nibnum, x, nodenum = 3, donenum = 0, level = 1;
    unsigned char nibs[17], curbyte;
    LRZ_NODE *root, *node;

    // Error checking
    if (chunk->srcoff >= chunk->srclen) return 1;

    // Get the number of nibbles for this tree
    curbyte = chunk->src[chunk->srcoff++];
    nibnum = (curbyte & 0x0F) + 2;
    if (nibnum == 2) return 0; // No nibbles for this tree
    if (chunk->srcoff + (nibnum >> 1) >= chunk->srclen)
        return 1; // Not enough bytes for this tree's definition

    // Process all of the nibbles for this tree
    for (x = 0; x < nibnum; x++) {
        if (x & 1) curbyte = chunk->src[chunk->srcoff++];
        else       curbyte >>= 4;
        nibs[x] = curbyte & 0x0F;
    }

    // Initialize the tree
    root = *out           = calloc(sizeof(LRZ_NODE), 1);
    node = root->child[0] = calloc(sizeof(LRZ_NODE), 1);
    root->child[1]        = calloc(sizeof(LRZ_NODE), 1);
    root->type = NODE_BRANCHING;
    root->child[0]->parent = root;
    root->child[1]->parent = root;

    // Construct the remainder of the tree
    while (!root->done) {

        // Check if this node needs further processing
        if (node->done) {
            node = node->parent;
            level--;
            continue;
        }

        // If a branching node, step down into the tree a level
        if (node->type == NODE_BRANCHING) {

            // Check if both child nodes are finished
            if (node->child[0]->done && node->child[1]->done) {
                node->done = 1;
                donenum++;
                continue;
            }

            // Step into the next unfinished child node
            if (node->child[0]->type == NODE_UNKNOWN)
                node = node->child[0];
            else node = node->child[1];
            level++;
            continue;
        }

        // If all the nibbles are less than the current level, return error
        for (x = 0; x < nibnum; x++)
            if (nibs[x] >= level) break;
        if (x == nibnum) return 1;

        // Try to find a value for this node
        for (x = 0; x < nibnum; x++)
            if (nibs[x] == level) break;

        // A value was found for this node
        if (x < nibnum) {
            nibs[x] = 0;
            node->type = NODE_VALUE;
            node->value = x;
            node->done = 1;
            donenum++;
            continue;
        }

        // No value could be found for this node; make it a branching node
        node->type = NODE_BRANCHING;
        node->child[0] = calloc(sizeof(LRZ_NODE), 1);
        node->child[1] = calloc(sizeof(LRZ_NODE), 1);
        node->child[0]->parent = node;
        node->child[1]->parent = node;
        nodenum += 2;
    }

    // If not all the nodes created were finalized, return error
    if (donenum != nodenum) return 1;

    // Return success
    return 0;
}

// Returns the value of the specified number of bits from the input stream
static int lrzdGetBits(LRZ_CHUNK *chunk, int count) {
    int ret;

    // Error checking
    if (chunk->bitsleft < count || count > MAX_BITS) return -1;
    chunk->bitsleft -= count;

    // Load the bits from the stream into the register
    while (count > chunk->subbits) {
        chunk->reg <<= chunk->subbits;
        if (chunk->srcoff < chunk->srclen)
            chunk->reg |= (int) chunk->src[chunk->srcoff++];
        count -= chunk->subbits;
        chunk->subbits = 8;
    }
    chunk->reg <<= count;
    chunk->subbits -= count;

    // Process the bits from the register
    ret = (chunk->reg & 0x00FFFF00) >> 8;
    chunk->reg &= 0x000000FF;
    return ret;
}

// Loads a number from input data encoded as a normal value
static int lrzdParse(LRZ_CHUNK *chunk) {
    int x = 1, count = 0;

    // Read (and count) bits until a 0 is encountered
    while (x) {
        x = lrzdGetBits(chunk, 1);
        if (x < 0) return -1; // Error reading bit
        count += x;
    }

    // Read that many additional bits from the input
    x = lrzdGetBits(chunk, count);
    if (x < 0) return -1; // Error reading bits

    // Return the composited value
    return (1 << count) - 1 + x;
}

// Loads a number from input data by traversing a binary tree
static int lrzdTraverse(LRZ_CHUNK *chunk, LRZ_NODE *node) {
    int x, ret;

    // Traverse the tree to a terminal node
    while (node->type == NODE_BRANCHING) {
        x = lrzdGetBits(chunk, 1);
        if (x < 0) return -1; // Error reading bit
        node = node->child[x];
    }

    // Return the number as a literal
    x = node->value;
    if (x < 2) return x;

    // Read additional bits from the data
    ret = 1 << --x;
    x = lrzdGetBits(chunk, x);
    if (x < 0) return -1; // Error reading bits
    return ret | x;
}

// This macro is used by lrzdChunk() to load a number from the input stream
#define GetNumber(var, tree) {           \
    if (tree != NULL)                    \
        var = lrzdTraverse(chunk, tree); \
    else var = lrzdParse(chunk);         \
}

// Processes a file chunk
static int lrzdChunk(LRZ_CHUNK *chunk) {
    int x, dist, len;

    // Initialize input/output variables
    chunk->srcoff = 0;
    chunk->dstoff = 0;

    // Process binary trees
    chunk->rawtree = chunk->lentree = chunk->disttree = NULL;
    if (lrzdGetTree(chunk, &chunk->rawtree))  return 1;
    if (lrzdGetTree(chunk, &chunk->lentree))  return 1;
    if (lrzdGetTree(chunk, &chunk->disttree)) return 1;

    // Initialize bit processing variables
    chunk->bitsleft = (chunk->srclen - chunk->srcoff) * 8;
    if (!chunk->bitsleft) return 1; // Insufficient data
    chunk->reg = (int) chunk->src[chunk->srcoff++];
    chunk->subbits = 8;

    // Process input data until all output data is written
    chunk->state = STATE_RAW;
    while (chunk->dstoff < chunk->dstlen) {

        // Read raw bytes from the input
        if (chunk->state == STATE_RAW) {

            // Read the number of bytes to copy
            GetNumber(x, chunk->rawtree);
            if (x < 0 || chunk->dstoff + x > chunk->dstlen)
                return LRZ_BADDATA;

            // Copy that many bytes
            for (len = x; len > 0; len--) {
                x = lrzdGetBits(chunk, 8);
                if (x < 0) return LRZ_BADDATA;
                chunk->dst[chunk->dstoff++] = 
                    (unsigned char) x;
            }
        }

        // Copy data with a distance/length pair
        else {

            // Read the length and distance values
            GetNumber(len,  chunk->lentree);
            GetNumber(dist, chunk->disttree);
            if (len < 0 || dist < 0) return LRZ_BADDATA;
            len += 2; dist++; // Implicit processing
            if (chunk->dstoff + len > chunk->dstlen || 
                chunk->dstoff - dist < 0) 
                return LRZ_BADDATA;

            // Copy the bytes from earlier in the output
            for (x = chunk->dstoff - dist; len > 0; len--)
                chunk->dst[chunk->dstoff++] = chunk->dst[x++];
        }

        // Alternate processing states
        chunk->state = 1 - chunk->state;
    }

    // Return success
    return 0;
}



////////////////////////////////////////////////////////////////////////////////
//                             Encoder Functions                              //
////////////////////////////////////////////////////////////////////////////////

// Locates the longest match of current input data in previous input data
// This implementation uses a linear brute force approach and is VERY slow
static int lrzeFindMatch(LRZ_CHUNK *chunk, int *d, int *l) {
    int maxlen = 0, maxdist = 0, len, dist;

    // Search for the longest (and closest) distance/length match
    for (dist = chunk->srcoff - 1; dist >= 0; dist--) {

        // Search for the first byte
        if (chunk->src[dist] != chunk->src[chunk->srcoff])
            continue;

        // Count how many bytes match
        for (len = 1; len < chunk->srclen - chunk->srcoff && 
            len < CLAMP; len++)
            if (chunk->src[dist + len] !=
                chunk->src[chunk->srcoff + len]) break;

        // Update the longest match if applicable
        if (len > maxlen) { maxlen = len; maxdist = chunk->srcoff - dist; }
    }

    // The maximum length was not long enough
    if (maxlen < 3) return 1;

    // Set return values
    *d = maxdist;
    *l = maxlen;
    return 0;
}

// Inserts bits into the output stream
static int lrzePutBits(LRZ_CHUNK *chunk, int bits, int count) {
    int shift;

    // Error checking
    if (count > MAX_BITS)        return LRZ_BADDATA;
    if (chunk->bitsleft < count) return LRZ_NOROOM;
    chunk->bitsleft -= count;

    // Prepare the shift register
    chunk->reg |= ((bits << (16 - count)) & 0xFFFF);

    // Output all the bits
    while (count) {

        // Get the number of bits that can be put into the next byte
        if (count > chunk->subbits)
            shift = chunk->subbits;
        else shift = count;

        // Process working variables
        chunk->reg <<= shift;
        chunk->subbits -= shift;
        count -= shift;

        // Output a byte if it's been filled up
        if (!chunk->subbits) {
            chunk->dst[chunk->dstoff++] = 
                (unsigned char) ((chunk->reg >> 16) & 0xFF);
            chunk->reg &= 0xFFFF;
            chunk->subbits = 8;
        }

    }

    // Return success
    return 0;
}

// Inserts a number into the output stream
static int lrzePutNumber(LRZ_CHUNK *chunk, int number) {
    int count, left, right, x;

    // Error checking
    if (number < 0 || number > CLAMP * 2) return LRZ_BADDATA;

    // Zero doesn't require any processing
    if (!number) return lrzePutBits(chunk, 0, 1);

    // Calculate how many bits are needed to express this number
    for (count = 1; count < MAX_BITS; count++)
        if ( ((1 << count) - 1) << 1 >= number) break;

    // Prepare and output values
    left = (1 << count) - 1;
    right = number - left;
    x = lrzePutBits(chunk,  left, count); if (x) return x;
    x = lrzePutBits(chunk,     0,     1); if (x) return x;
    x = lrzePutBits(chunk, right, count); if (x) return x;

    // Return success
    return 0;
}

// This macro is used by lrzeChunk() to encode raw bytes into the output
#define EncodeRaw() {                                           \
    if (lrzePutNumber(chunk, raws))                             \
        return LRZ_NOROOM;                                      \
    for (chunk->srcoff -= raws; raws > 0; raws--)               \
        if (lrzePutBits(chunk, chunk->src[chunk->srcoff++], 8)) \
        return LRZ_NOROOM;                                      \
}

// Creates a file chunk
static int lrzeChunk(LRZ_CHUNK *chunk) {
    int len, dist, raws;

    // Initialize encoder values
    chunk->srcoff   = chunk->dstoff = raws = 0;
    chunk->bitsleft = chunk->dstlen * 8;
    chunk->reg = 0;   chunk->subbits = 8;

    // Process all bytes of input
    while (chunk->srcoff < chunk->srclen) {

        // Attempt to find a long enough match for distance/length encoding
        if (lrzeFindMatch(chunk, &dist, &len)) {
            raws++;
            if (raws > CLAMP) return LRZ_BADDATA;
            chunk->srcoff++;
            continue;
        }

        // Encode raw bytes
        EncodeRaw();

        // Encode distance and length
        if (lrzePutNumber(chunk, len  - 2)) return LRZ_NOROOM;
        if (lrzePutNumber(chunk, dist - 1)) return LRZ_NOROOM;
        chunk->srcoff += len;
    }

    // Output any waiting data
    if (raws) EncodeRaw();
    if (chunk->subbits < 8) {
        if (lrzePutBits(chunk, 0, chunk->subbits))
        return LRZ_NOROOM;
    }

    // Return number of bytes encoded
    return chunk->dstoff;
}



////////////////////////////////////////////////////////////////////////////////
//                               API Functions                                //
////////////////////////////////////////////////////////////////////////////////

// Decompresses a compressed buffer
int lrzDecode(unsigned char *src, int srclen, unsigned char *dst, int dstlen) {
    int chunks, packed, unpacked, srcoff = 1, dstoff = 0, x;
    LRZ_CHUNK chunk;

    // Error checking
    if (src == NULL || dst == NULL || srclen < 9 || dstlen < 1)
        return LRZ_BADPARAMS;

    // Read the chunk count and check input size
    chunks = (int) src[0];
    if (!chunks || chunks * 5 + 1 > srclen) return LRZ_BADDATA;

    // Process all chunks
    for (; chunks > 0; chunks--, srcoff += packed, dstoff += unpacked) {

        // Process the chunk header
        packed   = ((int) src[srcoff + 1] << 8) | (int) src[srcoff];
        if (srcoff + packed > srclen) return LRZ_BADDATA;
        unpacked = ((int) src[srcoff + 3] << 8) | (int) src[srcoff + 2];
        if (dstoff + unpacked > dstlen) return LRZ_BADDATA;
        // src[srcoff + 4] seems to always be 4

        // Construct and process a chunk object
        chunk.src    = &src[srcoff + 5]; chunk.dst    = &dst[dstoff];
        chunk.srclen = packed - 5;       chunk.dstlen = unpacked;
        x = lrzdChunk(&chunk); lrzdFreeChunk(&chunk);
        if (x) return LRZ_BADDATA;
    }

    // Return success
    return LRZ_SUCCESS;
}

// Compresses a buffer to LRZ specifications
int lrzEncode(unsigned char *src, int srclen, unsigned char *dst, int dstlen) {
    int chunks, packed, unpacked, srcoff = 0, dstoff = 1;
    LRZ_CHUNK chunk;

    // Error checking
    if (src == NULL || dst == NULL || srclen < 1 || dstlen < 9)
        return LRZ_BADPARAMS;

    // Split the file into 0xF000-byte chunks
    chunks = srclen / 0xF000;
    if (srclen & 0x0FFF) chunks++;
    if (chunks > 255) return LRZ_TOOLARGE;
    dst[0] = (unsigned char) chunks;

    // Process all chunks
    for (; chunks > 0; chunks--, srcoff += unpacked, dstoff += packed) {

        // Initialize the header value
        unpacked = srclen - srcoff;
        if (unpacked > 0xF000) unpacked = 0xF000;

        // Construct and process a chunk object
        chunk.src    = &src[srcoff]; chunk.dst    = &dst[dstoff + 8];
        chunk.srclen = unpacked;     chunk.dstlen = dstlen - dstoff;
        packed = lrzeChunk(&chunk);
        if (packed < 0) return packed;
        packed = packed + 8;

        // Write chunk header
        dst[dstoff    ] = (unsigned char) ( packed         & 0xFF);
        dst[dstoff + 1] = (unsigned char) ((packed   >> 8) & 0xFF);
        dst[dstoff + 2] = (unsigned char) ( unpacked       & 0xFF);
        dst[dstoff + 3] = (unsigned char) ((unpacked >> 8) & 0xFF);
        dst[dstoff + 4] = 4; // Still not sure what this is
        dst[dstoff + 5] = 0; // Raw binary tree
        dst[dstoff + 6] = 0; // Length binary tree
        dst[dstoff + 7] = 0; // Distance binary tree
    }

    // Return number of bytes written
    return dstoff;
}

// Examines a compressed buffer to determine its decompressed size
int lrzGetSize(unsigned char *data, int len) {
    int chunks, packed, unpacked = 0, off = 1;

    // Error checking
    if (data == NULL || len < 1) return LRZ_BADPARAMS;

    // Read the chunk count and check input size
    chunks = (int) data[0];
    if (!chunks || chunks * 5 + 1 > len) return LRZ_BADDATA;

    // Check each chunk's packed size
    for (; chunks > 0; chunks--, off += packed) {
        packed = ((int) data[off + 1] << 8) | (int) data[off];
        if (off + packed > len) return LRZ_BADDATA;
        unpacked += (((int) data[off + 3] << 8) | (int) data[off + 2]);
    }

    // Return the accumulated unpacked size
    return unpacked;
}
